--with Reps.Ops, Rounding;
with Reps.Ops,Reps.IO;

pragma Elaborate_All (Reps.Ops);

package body Intervals is

  NegOne:     constant Rep := -One;
  NegTwo:     constant Rep := -Two;
  NegHalf:    constant Rep := -Half;
  Three:      constant Rep := Rep(3);
  Quarter:    constant Rep := Half/Two;
  OnePlus:    constant Rep := Rep(33)/Rep(32);
  Epsilon:    constant Rep := Half**99;

  Interval_Zero: constant Interval := (Zero,Zero);
  Interval_One:  constant Interval := (One,Zero);
  Interval_Two:  constant Interval := (Two,Zero);
  PlusMinus: constant Interval := (Zero,One);

  function IsNumeric(S: Interval) return Boolean is
  begin
    return False;
  end IsNumeric;

  function Err0(S: Interval) return Boolean is
  begin
    return S.R=Zero;
  end Err0;

  function Scal(K: Integer) return Interval is
  begin
    return (Rep(K),Zero);
  end Scal;

  function Scal(R: Rep) return Interval is
  begin
    return (R,Zero);
  end Scal;

  function Approx(S: Interval) return Rep is
  begin
    return S.C;
  end Approx;

  function Ball(R: Rep) return Interval is
  begin
    return (Zero,Abs(R));
  end Ball;

  function Ball(S: Interval) return Interval is
  begin
    return (Zero,Abs(S.C)+S.R);
  end Ball;

  function Scal(R1,R2: Rep) return Interval is
    Sum: constant Rep := R1+R2;
  begin
    if R1 >= R2 then
      return (Half*Sum,Half*((R1-R2)+(Sum+(-R1-R2))));
    end if;
    return (Half*Sum,Half*((R2-R1)+(Sum+(-R1-R2))));
  end Scal;

  function Center(S: Interval) return Interval is
  begin
    return (S.C,Zero);
  end Center;

  function Contains(S: Interval; Z: Rep) return Boolean is
  begin
     return (Sup(S)>=Z and Inf(S)<=Z);
  end Contains;

  procedure ResetCenter(S: in out Interval) is
  begin
    S.C := Zero;
  end ResetCenter;

  function ResetCenter(S: Interval) return Interval is
  begin
    return (Zero,S.R);
  end ResetCenter;

  procedure Split(S: in Interval; S0,SE: out Interval) is
  begin
    S0.C := S.C;
    S0.R := Zero;
    SE.C := Zero;
    SE.R := S.R;
  end Split;

  function Widen(S: Interval) return Interval is
-- only used in high level procedures
  begin
    return (S.C,S.R+Epsilon); -- avoid Gnat's Succ in Round_Up mode
  end Widen;

  function Inf(S: Interval) return Rep is
  begin
    return -(S.R-S.C);
  end Inf;

  function Sup(S: Interval) return Rep is
  begin
    return S.C+S.R;
  end Sup;

  pragma Inline (Inf,Sup);

  function Union(S1,S2: Interval) return Interval is
  begin
    return Scal(Rep'Min(Inf(S1),Inf(S2)),Rep'Max(Sup(S1),Sup(S2)));
  end Union;

  function "<"(S1,S2: Interval) return Boolean is
  begin
    if Sup(S1) < Inf(S2) then return True; end if;
    if Inf(S1) < Sup(S2) then raise Undefined; end if;
    return False;
  end "<";

  function "<="(S1,S2: Interval) return Boolean is
  begin
    if Sup(S1) <= Inf(S2) then return True; end if;
    if Inf(S1) <= Sup(S2) then raise Undefined; end if;
    return False;
  end "<=";

  function ">"(S1,S2: Interval) return Boolean is
  begin
    if Inf(S1) > Sup(S2) then return True; end if;
    if Sup(S1) > Inf(S2) then return True; end if;
    return False;
  end ">";

  function ">="(S1,S2: Interval) return Boolean is
  begin
    if Inf(S1) >= Sup(S2) then return True; end if;
    if Sup(S1) >= Inf(S2) then return True; end if;
    return False;
  end ">=";

  function Min(S1,S2: Interval) return Interval is
    U1: constant Rep := Sup(S1);
    L2: constant Rep := Inf(S2);
    L1,U2: Rep;
  begin
    if U1 <= L2 then
      return S1;
    end if;
    L1 := Inf(S1);
    U2 := Sup(S2);
    if U2 <= L1 then
      return S2;
    end if;
    return Scal(Rep'Min(L1,L2),Rep'Min(U1,U2));
  end Min;

  function Max(S1,S2: Interval) return Interval is
    U1: constant Rep := Sup(S1);
    L2: constant Rep := Inf(S2);
    L1,U2: Rep;
  begin
    if U1 <= L2 then
      return S2;
    end if;
    L1 := Inf(S1);
    U2 := Sup(S2);
    if U2 <= L1 then
      return S1;
    end if;
    return Scal(Rep'Max(L1,L2),Rep'Max(U1,U2));
   end Max;

  function Norm(S: Interval) return Interval is
    R: Rep;
  begin
    if S.C >= S.R then
      return S;
    elsif -S.C >= S.R then
      return (-S.C,S.R);
    end if;
    R := Half*(Abs(S.C)+S.R);
    return (R,R);
  end Norm;

  function InfAbs(S: Interval) return Rep is
  begin
    return -Rep'Min(S.R-Abs(S.C),Zero);
  end InfAbs;

  function SupAbs(S: Interval) return Rep is
  begin
    return Abs(S.C)+S.R;
  end SupAbs;

  function "-"(S: Interval) return Interval is
  begin
    return (-S.C,S.R);
  end "-";

  function "+"(S,T: Interval) return Interval is
    Sum: constant Rep := S.C+T.C;
  begin
    return (Sum,S.R+T.R+(Sum+(-S.C-T.C)));
  end "+";

  function "-"(S,T: Interval) return Interval is
    Diff: constant Rep := S.C-T.C;
  begin
    return (Diff,S.R+T.R+(Diff+(T.C-S.C)));
  end "-";

  function "*"(R: Rep; S: Interval) return Interval is
    Prod: constant Rep := R*S.C;
  begin
    return (Prod,Abs(R)*S.R+(Prod+(-R)*S.C));
  end "*";

  procedure ErrMult(R: in Radius; S: in out Interval) is
  begin
    S.R := R*S.R;
  end ErrMult;

  function "/"(S: Interval; R: Rep) return Interval is
    Quot: constant Rep := S.C/R;
  begin
    return (Quot,S.R/Abs(R)+(Quot+S.C/(-R)));
  end "/";

  function Sqr(S: Interval) return Interval is
    Prod: constant Rep := S.C*S.C;
    T: Interval;
  begin
    T:=(Prod,(Two*Abs(S.C)+S.R)*S.R+(Prod+S.C*(-S.C)));
    return Scal(Rep'Max(Zero,Inf(T)),Sup(T));
  end Sqr;

  function "*"(S,T: Interval) return Interval is
    Prod: constant Rep := S.C*T.C;
  begin
    return (Prod,Abs(S.C)*T.R+S.R*(Abs(T.C)+T.R)+(Prod+S.C*(-T.C)));
  end "*";

  function Inv(S: Interval) return Interval is
    Quot: constant Rep := One/S.C;
  begin
    if Abs(S.C) <= S.R then
      raise Constraint_Error;               -- division by zero
    end if;
    return (Quot,S.R/Abs(S.C)/(-(S.R-Abs(S.C)))+(Quot+NegOne/S.C));
  end Inv;

  function "/"(S,T: Interval) return Interval is
  begin
    return S*Inv(T);
  end "/";

  function "**"(S: Interval; I: Integer) return Interval is
  begin
    if I>0 then
      declare
        J: Integer := I;
        X,Y: Interval;
      begin
        X := S;
        if (J rem 2)=0 then
          Y := Interval_One;
        else
          Y := X;
        end if;
        J := J/2;
        while J>0 loop
          X := Sqr(X);
          if (J rem 2)>0 then
            Y := X*Y;
          end if;
          J := J/2;
        end loop;
        return Y;
      end;
    elsif I<0 then
      return Inv(S)**(-I);
    else
      return Interval_One;
    end if;
  end "**";

  procedure CoshSinh(S: in Interval; Sc,Ss: out Interval) is
    Iter: constant Integer := 6;
  begin
    if S = Interval_Zero then
      Sc := Interval_One;
      Ss := S;
    else
      declare
        K: Integer;
        R: Rep;
        X,T: Interval;
      begin
        K := 0;
        R := SupAbs(S);
        while R>Half loop
          K := K+1;
          R := R*Half;
        end loop;
        X := (Half**K)*S;
        T := Sqr(X);
        Sc := (One,Half);
        Ss := Sc;
        for I in reverse 1 .. Iter loop
          Sc := Interval_One+Sc*T/Rep((2*I)*(2*I-1));
          Ss := Interval_One+Ss*T/Rep((2*I)*(2*I+1));
        end loop;
        Ss := X*Ss;
        for I in 1 .. K loop
          T := Sc;
          Sc := Interval_One+Two*Sqr(Ss);
          Ss := Two*(T*Ss);
        end loop;
      end;
    end if;
  end CoshSinh;

  function Cosh(S: Interval) return Interval is
    Sc,Ss: Interval;
  begin
    CoshSinh(S,Sc,Ss);
    return Sc;
  end Cosh;

  function Sinh(S: Interval) return Interval is
    Sc,Ss: Interval;
  begin
    CoshSinh(S,Sc,Ss);
    return Ss;
  end Sinh;

  procedure CosSinn(S: in Interval; Sc,Ss: out Interval) is
    Iter: constant Integer := 6;
  begin
    if S = Interval_Zero then
      Sc := Interval_One;
      Ss := S;
    else
      declare
        K: Integer;
        R: Rep;
        X,T: Interval;
      begin
        K := 0;
        R := SupAbs(S);
        while R>Half loop
          K := K+1;
          R := R*Half;
        end loop;
        X := (Half**K)*S;
        T := Sqr(X);
        Sc := (One,Half);
        Ss := Sc;
        for I in reverse 1 .. Iter loop
          Sc := Interval_One+(Scal(-One)**I)*Sc*T/Rep((2*I)*(2*I-1));
          Ss := Interval_One+(Scal(-One)**I)*Ss*T/Rep((2*I)*(2*I+1));
        end loop;
        Ss := X*Ss;
        for I in 1 .. K loop
          T := Sc;
          Sc := Interval_One+Two*Sqr(Ss);
          Ss := Two*(T*Ss);
        end loop;
      end;
    end if;
  end CosSinn;


  procedure CosSin(S: in Interval; Sc,Ss: out Interval) is
      Iter: constant Integer := 8;
      Interval_Pi: constant Interval:=Scal(Rep(3.14159_26535_89793_23),Rep(3.14159_26535_89793_24));
      Co: Interval;
   begin
      if S = Interval_Zero then
         Sc := Interval_One;
         Ss := S;
      else
         declare
            K: Integer;
            R: Rep;
            X,T: Interval;
         begin
            K := 0;
            R := SupAbs(S);
            while R>Half loop
               K := K+1;
               R := R*Half;
            end loop;
            X := (Half**K)*S;
            T := Sqr(X);
            Sc := (One,Half);
            Ss := Sc;
            if ((Iter-1) mod 2)=0 then
               Co:=Interval_One;
            else
               Co:=-Interval_One;
            end if;
            for I in reverse 1 .. Iter loop
               Sc := Co+Sc*T/Rep((2*I)*(2*I-1));
               Ss := Co+Ss*T/Rep((2*I)*(2*I+1));
               Co:=-Co;
            end loop;

            Ss := X*Ss;
            for I in 1 .. K loop
               T := Sc;
               Sc := Interval_One-Two*Sqr(Ss);
               Ss := Two*(T*Ss);
            end loop;
         end;
      end if;
   end CosSin;



  function Cos(S: Interval) return Interval is
    Sc,Ss: Interval;
  begin
     CosSin(S,Sc,Ss);
    return Sc;
  end Cos;

  function Sin(S: Interval) return Interval is
    Sc,Ss: Interval;
  begin
     CosSin(S,Sc,Ss);
    return Ss;
  end Sin;


  function ExpUp(R: Rep) return Rep is
    Iter: constant Integer := 13;
    K: Integer := 0;
    X: Rep := R;
    Y: Rep := OnePlus;         --  33/32  assumes Iter>8
  begin
    if R<Zero then return One/ExpDown(-R); end if;
    if R=Zero then return One; end if;
    while X>Quarter loop
      K := K+1;
      X := Half*X;
    end loop;
    for I in reverse 1 .. Iter loop
      Y := One+X*Y/Rep(I);
    end loop;
    for I in 1 .. K loop
      Y := Y*Y;
    end loop;
    return Y;
  end ExpUp;

  function ExpDown(R: Rep) return Rep is
    Iter: constant Integer := 13;
    K: Integer := 0;
    X: Rep := R;
    Y: Rep := NegOne;
  begin
    if R<Zero then return -(NegOne/ExpUp(-R)); end if;
    if R=Zero then return One; end if;
    while X>Quarter loop
      K := K+1;
      X := Half*X;
    end loop;
    for I in reverse 1 .. Iter loop
      Y := NegOne+X*Y/Rep(I);
    end loop;
    for I in 1 .. K loop
      Y := (-Y)*Y;
    end loop;
    return -Y;
  end ExpDown;

  function Exp(S: Interval) return Interval is
  begin
    return Scal(ExpDown(Inf(S)),ExpUp(Sup(S)));
  end Exp;

  function Log(S: Interval) return Interval is
    use Reps.Ops;
  begin
    if S=Interval_One then return Interval_Zero; end if;
    declare
      Iter: constant Integer := 6;
      S0,LogS0,X,Y: Interval;
    begin
      LogS0 := Scal(Log(Approx(S)));   --- guess
      S0 := Exp(LogS0);
      X := (S0-S)/S0;                  --- Log(S) = Log(S0)+Log(1-X)
      Y := Interval_One-Scal(SupAbs(X));
      if Inf(Y) <= Zero then raise Constraint_Error; end if;
      Y := PlusMinus*Inv(Y)/Rep(Iter+1);
      for I in reverse 1 .. Iter loop
        Y := Inv(Scal(I))+X*Y;
      end loop;
      return LogS0-X*Y;
    end;
  end Log;

  procedure Ln(S1,S2: in out Interval)  is
     use Reps.Ops,Reps.IO;
  begin
     if S1=Interval_One and S2=Interval_Zero then
        S1:=Interval_Zero;
        S2:=Interval_Zero;
     end if;
     declare
        Iter: constant Integer := 6;
        S01,S02,LogS01,LogS02,X1,Y1,X2,Y2: Interval;
     begin
        --- guess
        LogS01 := Scal( Log( Sqrt(Sqr(Approx(S1))+Sqr(Approx(S2)) ) ) );
        LogS02 := -(Scal( ArcTan(Approx(S1)/Approx(S2)) )- Half*Scal(Rep(3.14159265358979323847))) ;
        if Inf(S2)<Zero then
           LogS02 := -( Scal( ArcTan(Approx(S1)/Approx(S2)) )+ Half*Scal(Rep(3.14159265358979323847)) );
        end if;



        S1:=LogS01;
        S2:=LogS02;

        return;

        S01 := Center(Exp(LogS01)*Cos(LogS02));
        S02 := Center(Exp(LogS01)*Sin(LogS02));

        --- Log(S) = Log(S0)+Log(1-X)

        X1 :=Interval_One-(S1*S01+S2*S02)/(Sqr(S01)+Sqr(S02) );
        X2 :=-(S2*S01-S1*S02) / (Sqr(S01)+Sqr(S02));
        Y1 := Interval_One -Scal(Sup(Sqrt(Sqr(X1)+Sqr(X2))));
        if Inf(Y1) <= Zero then raise Constraint_Error; end if;

        Y1 := PlusMinus*Inv(Y1)/Rep(Iter+1);
        --Y1 := Inv(Y1)/Rep(Iter+1);
        Y2 := Y1;

        for I in reverse 1 .. Iter loop
           Y1 := Inv(Scal(I))+X1*Y1-X2*Y2;
           Y2 := X1*Y2+Y1*X2;
        end loop;
        S1:=LogS01-(X1*Y1-X2*Y2);
        S2:=LogS02-(X1*Y2+X2*Y1);
     end;
  end Ln;

  function Sqrt(S: Interval) return Interval is
    use Reps.Ops;
  begin
    if S=Interval_Zero then return Interval_Zero; end if;
    if Inf(S)<Zero then raise Undefined; end if;
    declare
      S0,SqrtS0,X,Y: Interval;
    begin
      SqrtS0 := Scal(Sqrt(Approx(S))); --- guess
      S0 := Sqr(SqrtS0);
      X := (S-S0)/S0;                  --- Sqrt(S) = Sqrt(S0)*Sqrt(1+X)
      Y := (((Zero,One)*X-Scal(2))*X+Scal(8))*X/Scal(16);
      return SqrtS0*(Interval_One+Y);
    end;
  end Sqrt;

  function TruncExp(N: Natural; S: Interval) return Interval is
    Y: Interval := Interval_One;
  begin
    for I in reverse 1 .. N loop
      Y := Interval_One+(S*Y)/Rep(I);
    end loop;
    return Y;
  end TruncExp;

  function RemExp(N: Natural; S: Interval) return Interval is
--- Taylor remainder for Exp, divided by S**(N+1)/(N+1)!
    M: Integer;
    R: constant Rep := Sup(S);
    Y: Interval;
  begin
    if Inf(S)<Zero then
      raise Undefined;
    elsif R<Half then
      M := Integer'Max(N+1,17);
    elsif R<Two then
      M := Integer'Max(N+1,26);
    elsif R<Rep(8) then
      M := Integer'Max(N+1,46);
    elsif R<Rep(32) then
      M := Integer'Max(N+1,96);
    elsif R<Rep(128) then
      M := Integer'Max(N+1,245);
    elsif R<Rep(300) then
      M := Integer'Max(N+1,470);
    elsif R<Rep(512) then
      M := Integer'Max(N+1,730);
    elsif R<Rep(2048) then
      M := Integer'Max(N+1,1500);
    else
      raise Undefined;
    end if;
    Y := Inv(Interval_One-Scal(Zero,R)/Rep(M+1));
    for I in reverse N+2 .. M loop
      Y := Interval_One+(S*Y)/Rep(I);
    end loop;
    return Y;
  end RemExp;

--begin

-- Rounding.Restore_Checks;

end Intervals;
